# -*- coding: utf-8 -*-
import mock

import pytest

from s4 import sync
from s4.clients import SyncState, local, s3
from s4.sync import Resolution
from tests import utils


class TestGetStates(object):
    def test_empty_clients(self, s3_client, local_client):
        worker = sync.SyncWorker(s3_client, local_client)
        actual_output = list(worker.get_states())
        assert actual_output == []


class TestGetSyncStates(object):
    def test_empty(self, local_client, s3_client):
        assert sync.SyncWorker(local_client, s3_client).get_sync_states() == ({}, {})

    def test_correct_output(self, local_client, s3_client):
        utils.set_local_index(local_client, {
            'chemistry.txt': {
                'local_timestamp': 9431,
                'remote_timestamp': 9431,
            },
            'physics.txt': {
                'local_timestamp': 10000,
                'remote_timestamp': 10000,
            },
            'maltese.txt': {
                'local_timestamp': 7000,
                'remote_timestamp': 6000,
            },
        })
        utils.set_s3_index(s3_client, {
            'chemistry.txt': {
                'local_timestamp': 10000,
                'remote_timestamp': 9431,
            },
            'physics.txt': {
                'local_timestamp': 13000,
                'remote_timestamp': 12000,
            },
            'maltese.txt': {
                'local_timestamp': 6000,
                'remote_timestamp': 6000,
            },
        })

        utils.set_local_contents(local_client, 'history.txt', timestamp=5000)
        utils.set_s3_contents(s3_client, 'art.txt', timestamp=200000)
        utils.set_local_contents(local_client, 'english.txt', timestamp=90000)
        utils.set_s3_contents(s3_client, 'english.txt', timestamp=93000)
        utils.set_s3_contents(s3_client, 'chemistry.txt', timestamp=10000)
        utils.set_local_contents(local_client, 'physics.txt', timestamp=11000)
        utils.set_s3_contents(s3_client, 'physics.txt', timestamp=13000)
        utils.set_local_contents(local_client, 'maltese.txt', timestamp=7000)
        utils.set_s3_contents(s3_client, 'maltese.txt', timestamp=8000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            'english.txt': (
                SyncState(SyncState.CREATED, 90000, None),
                SyncState(SyncState.CREATED, 93000, None)
            ),
            'physics.txt': (
                SyncState(SyncState.UPDATED, 11000, 10000),
                SyncState(SyncState.NOCHANGES, 13000, 12000),
            )
        }
        expected_resolutions = {
            'maltese.txt': Resolution(
                Resolution.UPDATE, local_client, s3_client, 'maltese.txt', 8000
            ),
            'chemistry.txt': Resolution(
                Resolution.DELETE, s3_client, None, 'chemistry.txt', 9431
            ),
            'history.txt': Resolution(
                Resolution.CREATE, s3_client, local_client, 'history.txt', 5000
            ),
            'art.txt': Resolution(
                Resolution.CREATE, local_client, s3_client, 'art.txt', 200000
            ),
        }
        assert unhandled_events == expected_unhandled_events
        assert resolutions == expected_resolutions

    def test_nochanges_but_different_remote_timestamps(self, local_client, s3_client):
        utils.set_local_index(local_client, {
            'german.txt': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            }
        })
        utils.set_s3_index(s3_client, {
            'german.txt': {
                'local_timestamp': 6000,
                'remote_timestamp': 6000,
            }
        })
        utils.set_local_contents(local_client, 'german.txt', timestamp=4000)
        utils.set_s3_contents(s3_client, 'german.txt', timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_resolutions = {
            'german.txt': Resolution(
                "UPDATE", local_client, s3_client, 'german.txt', 6000)
        }
        assert resolutions == expected_resolutions
        assert unhandled_events == {}

    def test_updated_but_different_remote_timestamp(self, local_client, s3_client):
        utils.set_local_index(local_client, {
            'biology.txt': {
                'local_timestamp': 4000,
                'remote_timestamp': 3000,
            }
        })
        utils.set_s3_index(s3_client, {
            'biology.txt': {
                'local_timestamp': 6000,
                'remote_timestamp': 6000,
            }
        })
        utils.set_local_contents(local_client, 'biology.txt', timestamp=4500)
        utils.set_s3_contents(s3_client, 'biology.txt', timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            'biology.txt': (
                SyncState(SyncState.UPDATED, 4500, 3000), SyncState(SyncState.NOCHANGES, 6000, 6000)
            )
        }
        assert resolutions == {}
        assert unhandled_events == expected_unhandled_events

    def test_deleted_but_different_remote_timestamp(self, local_client, s3_client):
        utils.set_local_index(local_client, {
            'chemistry.txt': {
                'local_timestamp': 4000,
                'remote_timestamp': 3000,
            }
        })
        utils.set_s3_index(s3_client, {
            'chemistry.txt': {
                'local_timestamp': 6000,
                'remote_timestamp': 6000,
            }
        })
        utils.set_s3_contents(s3_client, 'chemistry.txt', timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            'chemistry.txt': (
                SyncState(SyncState.DELETED, None, 3000), SyncState(SyncState.NOCHANGES, 6000, 6000)
            )
        }
        assert resolutions == {}
        assert unhandled_events == expected_unhandled_events

    def test_deleted_doesnotexist(self, local_client, s3_client):
        utils.set_local_index(local_client, {
            'physics.txt': {
                'local_timestamp': 5000,
                'remote_timestamp': 4550,
            }
        })
        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        assert resolutions == {}
        assert unhandled_events == {}


def assert_contents(clients, key, data=None, timestamp=None):
    for client in clients:
        sync_object = client.get(key)
        if data is not None:
            assert sync_object.fp.read() == data
        if timestamp is not None:
            assert sync_object.timestamp == timestamp


def assert_remote_timestamp(clients, key, expected_timestamp):
    for client in clients:
        assert client.get_remote_timestamp(key) == expected_timestamp


def assert_local_keys(clients, expected_keys):
    for client in clients:
        assert sorted(client.get_local_keys()) == sorted(expected_keys)


def assert_existence(clients, keys, exists):
    for client in clients:
        for key in keys:
            assert (client.get(key) is not None) == exists


class TestSyncWorker(object):
    def test_repr(self):
        local_client = local.LocalSyncClient('/home/bobs/burgers')
        s3_client = s3.S3SyncClient(None, 'burgerbucket', 'foozie')
        worker = sync.SyncWorker(local_client, s3_client)
        assert repr(worker) == 'SyncWorker</home/bobs/burgers/, s3://burgerbucket/foozie/>'


def TestGetResolution(object):
    def test_unknown(self, local_client, s3_client):
        state = SyncState("unkonwn state", None, None)
        with pytest.raises(ValueError):
            sync.get_resolution("test", state, local_client, s3_client)


class TestIntegrations(object):
    def test_dry_run(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, 'colors/cream', 9999, '#ddeeff')

        utils.set_local_contents(local_client, 'colors/red', 5000, '#ff0000')
        utils.set_local_contents(local_client, 'colors/green', 3000, '#00ff00')
        utils.set_local_contents(local_client, 'colors/blue', 2000, '#0000ff')

        local_client.put = mock.MagicMock()
        s3_client.put = mock.MagicMock()

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(dry_run=True)

        assert not local_client.put.called
        assert not s3_client.put.called

    def test_action_callback(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, 'colors/cream', 9999)

        utils.set_local_contents(local_client, 'colors/green', 3000)
        utils.set_local_contents(local_client, 'colors/blue', 2000)

        callback_mock = mock.MagicMock()

        worker = sync.SyncWorker(
            local_client,
            s3_client,
            action_callback=callback_mock
        )
        worker.sync()

        assert callback_mock.call_count == 3
        assert callback_mock.call_args_list[0][0][0].key == 'colors/blue'
        assert callback_mock.call_args_list[1][0][0].key == 'colors/cream'
        assert callback_mock.call_args_list[2][0][0].key == 'colors/green'

    def test_local_with_s3(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, 'colors/cream', 9999, '#ddeeff')

        utils.set_local_contents(local_client, 'colors/red', 5000, '#ff0000')
        utils.set_local_contents(local_client, 'colors/green', 3000, '#00ff00')
        utils.set_local_contents(local_client, 'colors/blue', 2000, '#0000ff')

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        expected_keys = [
            'colors/red',
            'colors/green',
            'colors/blue',
            'colors/cream'
        ]
        assert_local_keys(clients, expected_keys)
        assert_contents(clients, 'colors/red', b'#ff0000')
        assert_contents(clients, 'colors/green', b'#00ff00')
        assert_contents(clients, 'colors/blue', b'#0000ff')
        assert_contents(clients, 'colors/cream', b'#ddeeff')
        assert_remote_timestamp(clients, 'colors/red', 5000)
        assert_remote_timestamp(clients, 'colors/green', 3000)
        assert_remote_timestamp(clients, 'colors/blue', 2000)
        assert_remote_timestamp(clients, 'colors/cream', 9999)

        utils.delete_local(local_client, 'colors/red')

        worker.sync()
        expected_keys = [
            'colors/green',
            'colors/blue',
            'colors/cream'
        ]
        assert_local_keys(clients, expected_keys)

    def test_fresh_sync(self, local_client, s3_client):
        utils.set_local_contents(local_client, 'foo', timestamp=1000)
        utils.set_local_contents(local_client, 'bar', timestamp=2000)
        utils.set_s3_contents(s3_client, 'baz', timestamp=3000, data='what is up?')

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        assert_local_keys(clients, ['foo', 'bar', 'baz'])
        assert_remote_timestamp(clients, 'foo', 1000)
        assert_remote_timestamp(clients, 'bar', 2000)
        assert_remote_timestamp(clients, 'baz', 3000)
        assert_existence(clients, ['foo', 'bar', 'baz'], True)
        assert_contents(clients, 'baz', b'what is up?')

        utils.delete_local(local_client, 'foo')
        utils.set_local_contents(local_client, 'test', timestamp=5000)
        utils.set_s3_contents(s3_client, 'hello', timestamp=6000)
        utils.set_s3_contents(s3_client, 'baz', timestamp=8000, data='just syncing some stuff')

        worker.sync()

        assert_existence(clients, ['foo'], False)
        assert_local_keys(clients, ['test', 'bar', 'baz', 'hello'])
        assert_remote_timestamp(clients, 'bar', 2000)
        assert_remote_timestamp(clients, 'foo', 1000)
        assert_remote_timestamp(clients, 'test', 5000)
        assert_remote_timestamp(clients, 'hello', 6000)
        assert_remote_timestamp(clients, 'baz', 8000)
        assert_contents(clients, 'baz', b'just syncing some stuff')

        worker.sync()

    def test_three_way_sync(self, local_client, s3_client, local_client_2):
        utils.set_local_contents(local_client, 'foo', timestamp=1000)
        utils.set_s3_contents(s3_client, 'bar', timestamp=2000, data='red')
        utils.set_local_contents(local_client_2, 'baz', timestamp=3000)

        worker_1 = sync.SyncWorker(local_client, s3_client)
        worker_2 = sync.SyncWorker(local_client_2, s3_client)
        worker_3 = sync.SyncWorker(local_client, local_client_2)

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        clients = [local_client, s3_client, local_client_2]

        assert_local_keys(clients, ['foo', 'bar', 'baz'])
        assert_contents(clients, 'bar', b'red')
        assert_remote_timestamp(clients, 'foo', 1000)
        assert_remote_timestamp(clients, 'bar', 2000)
        assert_remote_timestamp(clients, 'baz', 3000)

        utils.set_s3_contents(s3_client, 'bar', timestamp=8000, data='green')

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        assert_local_keys(clients, ['foo', 'bar', 'baz'])
        assert_contents(clients, 'bar', b'green')
        assert_remote_timestamp(clients, 'foo', 1000)
        assert_remote_timestamp(clients, 'bar', 8000)
        assert_remote_timestamp(clients, 'baz', 3000)

        utils.delete_local(local_client_2, 'foo')

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        assert_existence(clients, ['foo'], False)
        assert_local_keys(clients, ['bar', 'baz'])
        assert_remote_timestamp(clients, 'foo', 1000)

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

    def test_ignore_conflicts(self, local_client, s3_client):
        utils.set_local_contents(local_client, 'foo', timestamp=2000)
        utils.set_s3_contents(s3_client, 'foo', timestamp=3000)
        utils.set_s3_contents(s3_client, 'bar', timestamp=5600, data='usador')

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice='ignore')

        clients = [local_client, s3_client]
        assert_local_keys(clients, ['foo', 'bar'])
        assert_remote_timestamp(clients, 'foo', None)
        assert_remote_timestamp(clients, 'bar', 5600)
        assert_existence(clients, ['foo', 'bar'], True)
        assert_contents(clients, 'bar', b'usador')

        # Check that none of the files have been modified due to ignore flag
        assert local_client.get_real_local_timestamp('foo') == 2000
        assert s3_client.get_real_local_timestamp('foo') == 3000

    def test_conflict_choose_first_client(self, local_client, s3_client):
        utils.set_local_contents(local_client, 'foo', timestamp=2000, data='abc')
        utils.set_s3_contents(s3_client, 'foo', timestamp=3000)

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice='1')

        clients = [local_client, s3_client]
        assert_local_keys(clients, ['foo'])
        # Chooses first client
        assert_remote_timestamp(clients, 'foo', 2000)
        assert_contents(clients, 'foo', b'abc')

    def test_conflict_choose_second_client(self, local_client, s3_client):
        utils.set_local_contents(local_client, 'foo', timestamp=2000)
        utils.set_s3_contents(s3_client, 'foo', timestamp=3000, data='123')

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice='2')

        clients = [local_client, s3_client]
        assert_local_keys(clients, ['foo'])
        # Chooses second client
        assert_remote_timestamp(clients, 'foo', 3000)
        assert_contents(clients, 'foo', b'123')

    def test_previously_deleted_now_created(self, local_client, s3_client):
        utils.set_s3_index(s3_client, {
            'foo': {
                'local_timestamp': None,
                'remote_timestamp': 4000,
            }
        })
        utils.set_local_index(local_client, {
            'foo': {
                'local_timstamp': None,
                'remote_timestamp': 4000,
            }
        })
        utils.set_local_contents(local_client, 'foo', timestamp=7000)

        # Will create previously deleted file
        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        assert_local_keys(clients, ['foo'])
        assert_remote_timestamp(clients, 'foo', 7000)

        # delete the file again and check that it is successful
        utils.delete_local(local_client, 'foo')

        worker.sync()
        assert_local_keys(clients, [])
        assert_remote_timestamp(clients, 'foo', 7000)


class TestRunResolutions(object):
    def test_empty(self, local_client, s3_client):
        worker = sync.SyncWorker(local_client, s3_client)
        assert worker.run_resolutions({}) == []

    def test_invalid_action(self, local_client, s3_client):
        worker = sync.SyncWorker(local_client, s3_client)
        with pytest.raises(ValueError):
            worker.run_resolutions({
                'foo': Resolution('UNKNOWN', None, None, None, None),
            })

    def test_correct_output(self, local_client, s3_client):
        def keyboard_interrupt():
            raise KeyboardInterrupt()

        clients = [local_client, s3_client]
        worker = sync.SyncWorker(local_client, s3_client)

        utils.set_local_contents(local_client, 'foo')
        utils.set_s3_contents(s3_client, 'baz', timestamp=2000, data='testing')
        success = worker.run_resolutions({
            'foo': Resolution(Resolution.DELETE, local_client, None, 'foo', 1000),
            'baz': Resolution(Resolution.CREATE, local_client, s3_client, 'baz', 20),
        })
        assert sorted(success) == sorted(['foo', 'baz'])
        assert_local_keys(clients, ['baz'])


class TestMoveClient(object):
    def test_correct_behaviour(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, 'art.txt', data='swirly abstract objects')

        worker = sync.SyncWorker(local_client, s3_client)

        resolution = Resolution(
            action=Resolution.CREATE,
            to_client=local_client,
            from_client=s3_client,
            key='art.txt',
            timestamp=6000,
        )
        worker.move_client(resolution)

        assert utils.get_local_contents(local_client, 'art.txt') == b'swirly abstract objects'
        assert local_client.get_remote_timestamp('art.txt') == 6000



# -*- coding: utf-8 -*-

import argparse
import logging
import sys

from s4 import VERSION
from s4 import utils
from s4.commands.add_command import AddCommand
from s4.commands.daemon_command import DaemonCommand
from s4.commands.edit_command import EditCommand
from s4.commands.ls_command import LsCommand
from s4.commands.rm_command import RmCommand
from s4.commands.sync_command import SyncCommand
from s4.commands.targets_command import TargetsCommand


def main(arguments):
    parser = argparse.ArgumentParser(
        prog='s4',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            'Fast and cheap synchronisation of files with Amazon S3\n'
            '\n'
            'Version: {}\n'
            '\n'
            'To start off, add a Target with the "add" command\n'
        ).format(VERSION),
    )
    parser.add_argument(
        '--log-level',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
    )
    parser.add_argument(
        '--no-colors',
        action='store_true',
        help='Display without colors',
    )
    parser.add_argument(
        '--timestamps',
        action='store_true',
        help='Display timestamps for each log message',
    )
    subparsers = parser.add_subparsers(dest='command')

    daemon_parser = subparsers.add_parser('daemon', help="Run S4 sync continiously")
    daemon_parser.add_argument('targets', nargs='*')
    daemon_parser.add_argument('--read-delay', default=1000, type=int)
    daemon_parser.add_argument('--conflicts', default='ignore', choices=['1', '2', 'ignore'])

    add_parser = subparsers.add_parser('add', help="Add a new Target to synchronise")
    add_parser.add_argument(
        '--copy-target-credentials',
        '-C',
        help="Copy credentials from an existing target instead of typing them in again"
    )

    sync_parser = subparsers.add_parser('sync', help="Synchronise Targets with S3")
    sync_parser.add_argument('targets', nargs='*')
    sync_parser.add_argument('--conflicts', default=None, choices=['1', '2', 'ignore'])
    sync_parser.add_argument('--dry-run', action='store_true')

    edit_parser = subparsers.add_parser('edit', help="Edit Target details")
    edit_parser.add_argument('target')

    subparsers.add_parser('targets', help="Print available Targets")

    subparsers.add_parser('version', help="Print S4 Version")

    ls_parser = subparsers.add_parser('ls', help="Display list of files for a Target")
    ls_parser.add_argument('target')
    ls_parser.add_argument('--sort-by', '-s', choices=['key', 'local', 's3'], default='key')
    ls_parser.add_argument('--descending', '-d', action='store_true')
    ls_parser.add_argument(
        '--all', '-A',
        dest='show_all',
        action='store_true',
        help='show deleted files',
    )

    remove_parser = subparsers.add_parser('rm', help="Remove a Target")
    remove_parser.add_argument('target')

    args = parser.parse_args(arguments)

    if args.log_level == 'DEBUG':
        log_format = '%(levelname)s:%(module)s:%(lineno)s %(message)s'
    else:
        log_format = '%(message)s'

    if args.timestamps:
        log_format = '%(asctime)s: ' + log_format

    logging.basicConfig(format=log_format, level=args.log_level)

    # shut boto up
    logging.getLogger('boto3').setLevel(logging.CRITICAL)
    logging.getLogger('botocore').setLevel(logging.CRITICAL)
    logging.getLogger('nose').setLevel(logging.CRITICAL)
    logging.getLogger('s3transfer').setLevel(logging.CRITICAL)
    logging.getLogger('filelock').setLevel(logging.CRITICAL)

    logger = logging.getLogger(__name__)
    logger.setLevel(args.log_level)

    config = utils.get_config()

    try:
        command = None
        if args.command == 'version':
            print(VERSION)
            return
        elif args.command == 'sync':
            command = SyncCommand(args, config, logger)
        elif args.command == 'targets':
            command = TargetsCommand(args, config, logger)
        elif args.command == 'add':
            command = AddCommand(args, config, logger)
        elif args.command == 'edit':
            command = EditCommand(args, config, logger)
        elif args.command == 'ls':
            command = LsCommand(args, config, logger)
        elif args.command == 'rm':
            command = RmCommand(args, config, logger)
        elif args.command == 'daemon':
            command = DaemonCommand(args, config, logger)

        if command:
            try:
                command.run()
            except Exception as e:
                logger.error('An unhandled error has occurred: %s', e)
                # Only display a scary stack trace to the user if in DEBUG mode
                if args.log_level == 'DEBUG':
                    raise e
        else:
            parser.print_help()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main(sys.argv[1:])



# -*- encoding: utf-8 -*-

from s4.commands.targets_command import TargetsCommand

from tests.utils import create_logger


class TestTargetsCommand(object):

    def test_empty(self, capsys):
        command = TargetsCommand(None, {'targets': {}}, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == err == ''

    def test_correct_output(self, capsys):
        config = {
            'targets': {
                'Personal': {
                    's3_uri': 's3://mybackup/Personal',
                    'local_folder': '/home/user/Documents',
                },
                'Studies': {
                    's3_uri': 's3://something/something/Studies',
                    'local_folder': '/media/backup/Studies',
                },
            }
        }

        command = TargetsCommand(None, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert err == ''
        assert out == (
            'Personal: [/home/user/Documents <=> s3://mybackup/Personal]\n'
            'Studies: [/media/backup/Studies <=> s3://something/something/Studies]\n'
        )



#! -*- encoding: utf8 -*-
import os

from s4 import utils
from s4.commands import Command


class EditCommand(Command):
    def run(self):
        if 'targets' not in self.config:
            self.logger.info('You have not added any targets yet')
            self.logger.info('Use the "add" command to do this')
            return

        if self.args.target not in self.config['targets']:
            all_targets = sorted(list(self.config['targets'].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info('Choices are: %s', all_targets)
            return

        entry = self.config['targets'][self.args.target]

        local_folder = entry.get('local_folder', '')
        s3_uri = entry.get('s3_uri', '')
        aws_access_key_id = entry.get('aws_access_key_id')
        aws_secret_access_key = entry.get('aws_secret_access_key')
        region_name = entry.get('region_name')

        new_local_folder = utils.get_input('local folder [{}]: '.format(local_folder))
        new_s3_uri = utils.get_input('s3 uri [{}]: '.format(s3_uri))
        new_aws_access_key_id = utils.get_input(
            'AWS Access Key ID [{}]: '.format(aws_access_key_id)
        )

        secret_key_prompt = 'AWS Secret Access Key [{}]: '.format(aws_secret_access_key)
        new_aws_secret_access_key = utils.get_input(secret_key_prompt, secret=True)

        new_region_name = utils.get_input('region name [{}]: '.format(region_name))

        if new_local_folder:
            entry['local_folder'] = os.path.expanduser(new_local_folder)
        if new_s3_uri:
            entry['s3_uri'] = new_s3_uri
        if new_aws_access_key_id:
            entry['aws_access_key_id'] = new_aws_access_key_id
        if new_aws_secret_access_key:
            entry['aws_secret_access_key'] = new_aws_secret_access_key
        if new_region_name:
            entry['region_name'] = new_region_name

        self.config['targets'][self.args.target] = entry

        utils.set_config(self.config)



# -*- encoding: utf-8 -*-
import argparse
import json

import mock

from s4.commands.edit_command import EditCommand

from tests import utils


@mock.patch('s4.utils.get_input')
class TestEditCommand(object):
    def test_no_targets(self, get_input, capsys):
        command = EditCommand(None, {}, utils.create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'You have not added any targets yet\n'
            'Use the "add" command to do this\n'
        )

    def test_missing_target(self, get_input, capsys):
        args = argparse.Namespace(target='idontexist')
        config = {
            'targets': {'foo': {}}
        }
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            '"idontexist" is an unknown target\n'
            'Choices are: [\'foo\']\n'
        )

    def test_no_changes(self, get_input, config_file):
        fake_stream = utils.FakeInputStream([
            '',
            '',
            '',
            '',
            '',
        ])
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target='foo')
        config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        with open(config_file, 'r') as fp:
            config = json.load(fp)

        expected_config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        assert expected_config == config

    def test_correct_output(self, get_input, config_file):
        fake_stream = utils.FakeInputStream([
            '/home/user/Documents',
            's3://buckets/mybackup222',
            '9999999999',
            'bbbbbbbbbbbbbbbbbbbbbbbb',
            'eu-west-2',
        ])
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target='foo')
        config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        with open(config_file, 'r') as fp:
            config = json.load(fp)

        expected_config = {
            'targets': {'foo': {
                'local_folder': '/home/user/Documents',
                's3_uri': 's3://buckets/mybackup222',
                'aws_access_key_id': '9999999999',
                'aws_secret_access_key': 'bbbbbbbbbbbbbbbbbbbbbbbb',
                'region_name': 'eu-west-2',
            }}
        }
        assert expected_config == config



# -*- coding: utf-8 -*-

import logging
import os
import shutil
import tempfile
import uuid

import boto3

from faker import Faker

import mock

import moto

import pytest

from s4.clients.local import LocalSyncClient
from s4.clients.s3 import S3SyncClient

fake = Faker()


@pytest.fixture
def a_logger():
    result = logging.getLogger(str(uuid.uuid4()))
    result.setLevel(logging.INFO)
    result.handlers = []
    result.addHandler(logging.StreamHandler())
    return result


@pytest.fixture
def config_file():
    fd, temp_path = tempfile.mkstemp()
    mocker = mock.patch('s4.utils.CONFIG_FILE_PATH', temp_path)
    mocker.start()
    yield temp_path
    mocker.stop()
    os.close(fd)
    os.remove(temp_path)


@pytest.fixture
def local_client():
    folder = tempfile.mkdtemp()
    yield LocalSyncClient(folder)
    shutil.rmtree(folder)


@pytest.fixture
def local_client_2():
    folder = tempfile.mkdtemp()
    yield LocalSyncClient(folder)
    shutil.rmtree(folder)


@pytest.fixture
def s3_client():
    mock = moto.mock_s3()
    mock.start()
    boto_client = boto3.client(
        's3',
        aws_access_key_id='',
        aws_secret_access_key='',
        aws_session_token='',
    )
    bucket_name = fake.user_name()
    boto_client.create_bucket(Bucket=bucket_name)
    yield S3SyncClient(boto_client, bucket_name, fake.uri_path())
    mock.stop()



#! -*- encoding: utf8 -*-

import mock

from s4 import diff
from tests import utils


class TestShowDiff(object):
    @mock.patch('shutil.which')
    def test_diff_not_found(self, which, capsys, local_client, s3_client):
        which.return_value = None
        diff.show_diff(local_client, s3_client, "something")

        out, err = capsys.readouterr()
        assert out == (
            'Missing required "diff" executable.\n'
            "Install this using your distribution's package manager\n"
        )

    @mock.patch('shutil.which')
    def test_less_not_found(self, which, capsys, local_client, s3_client):
        def missing_less(value):
            return None if value == 'less' else 'something'

        which.side_effect = missing_less
        diff.show_diff(local_client, s3_client, "something")

        out, err = capsys.readouterr()
        assert out == (
            'Missing required "less" executable.\n'
            "Install this using your distribution's package manager\n"
        )

    @mock.patch('subprocess.call')
    def test_diff(self, call, local_client, s3_client):
        utils.set_local_contents(local_client, "something", 4000, "wow")
        utils.set_s3_contents(s3_client, "something", 3000, "nice")

        diff.show_diff(local_client, s3_client, "something")

        assert call.call_count == 2
        assert call.call_args_list[0][0][0][0] == "diff"
        assert call.call_args_list[1][0][0][0] == "less"



# -*- encoding: utf-8 -*-

import argparse
import json

import mock

from s4.commands.add_command import AddCommand
from tests import utils


@mock.patch('s4.utils.get_input')
class TestAddCommand(object):
    def test_correct_behaviour(self, get_input, config_file):
        fake_stream = utils.FakeInputStream([
            '/home/user/Documents',
            's3://mybucket/Documents',
            'eu-west-2',
            'aaaaaaaaaaaaaaaaaaaaaaaa',
            'bbbbbbbbbbbbbbbbbbbbbbbb',
            '',
        ])
        get_input.side_effect = fake_stream
        args = argparse.Namespace(copy_target_credentials=None)

        command = AddCommand(args, {'targets': {}}, utils.create_logger())
        command.run()

        with open(config_file, 'r') as fp:
            new_config = json.load(fp)

        expected_config = {
            'targets': {
                'Documents': {
                    'local_folder': '/home/user/Documents',
                    's3_uri': 's3://mybucket/Documents',
                    'aws_access_key_id': 'aaaaaaaaaaaaaaaaaaaaaaaa',
                    'aws_secret_access_key': 'bbbbbbbbbbbbbbbbbbbbbbbb',
                    'region_name': 'eu-west-2'
                }
            }
        }
        assert new_config == expected_config

    def test_copy_target_credentials(self, get_input, config_file):
        fake_stream = utils.FakeInputStream([
            '/home/user/Animals',
            's3://mybucket/Zoo',
            'us-west-2',
            'Beasts',
        ])
        get_input.side_effect = fake_stream
        args = argparse.Namespace(copy_target_credentials='bar')

        command = AddCommand(
            args,
            {
                'targets': {
                    'bar': {
                        'aws_secret_access_key': 'bar-secretz',
                        'aws_access_key_id': 'so-much-bar',
                    }
                }
            },
            utils.create_logger()
        )
        command.run()

        with open(config_file, 'r') as fp:
            new_config = json.load(fp)

        expected_config = {
            'targets': {
                'bar': {
                    'aws_access_key_id': 'so-much-bar',
                    'aws_secret_access_key': 'bar-secretz',
                },
                'Beasts': {
                    'local_folder': '/home/user/Animals',
                    's3_uri': 's3://mybucket/Zoo',
                    'aws_access_key_id': 'so-much-bar',
                    'aws_secret_access_key': 'bar-secretz',
                    'region_name': 'us-west-2'
                }
            }
        }
        assert new_config == expected_config

    def test_copy_target_credentials_bad_target(self, get_input, capsys):
        fake_stream = utils.FakeInputStream([
            '/home/user/Animals',
            's3://mybucket/Zoo',
            'us-west-2',
            'Beasts',
        ])
        get_input.side_effect = fake_stream
        args = argparse.Namespace(copy_target_credentials='Foo')

        command = AddCommand(args, {'targets': {'bar': {}}}, utils.create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            '"Foo" is an unknown target\n'
            'Choices are: [\'bar\']\n'
        )

    def test_custom_target_name(self, get_input, config_file):
        fake_stream = utils.FakeInputStream([
            '/home/user/Music',
            's3://mybucket/Musiccccc',
            'us-west-1',
            '1234567890',
            'abcdefghij',
            'Tunes',
        ])
        get_input.side_effect = fake_stream
        args = argparse.Namespace(copy_target_credentials=None)

        command = AddCommand(args, {'targets': {}}, utils.create_logger())
        command.run()

        with open(config_file, 'r') as fp:
            new_config = json.load(fp)

        expected_config = {
            'targets': {
                'Tunes': {
                    'local_folder': '/home/user/Music',
                    's3_uri': 's3://mybucket/Musiccccc',
                    'aws_access_key_id': '1234567890',
                    'aws_secret_access_key': 'abcdefghij',
                    'region_name': 'us-west-1'
                }
            }
        }
        assert new_config == expected_config



# -*- encoding: utf-8 -*-

import argparse

from inotify_simple import Event, flags

import mock

import pytest

from s4.commands.daemon_command import DaemonCommand
from tests.utils import create_logger


class FakeINotify(object):
    def __init__(self, events, wd_map):
        self.events = events
        self.wd_map = wd_map

    def add_watches(self, *args, **kwargs):
        return self.wd_map

    def read(self, *args, **kwargs):
        return self.events


@mock.patch('s4.sync.SyncWorker')
@mock.patch('s4.commands.daemon_command.INotifyRecursive')
class TestDaemonCommand(object):
    def single_term(self, index):
        """Simple terminator for the daemon command"""
        return index >= 1

    @mock.patch('s4.commands.daemon_command.supported', False)
    def test_os_not_supported(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts='ignore', read_delay=0)

        command = DaemonCommand(args, {}, create_logger())
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'Cannot run INotify on your operating system\n'
            'Only Linux machines are officially supported for this command\n'
        )
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_no_targets(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts='ignore', read_delay=0)

        command = DaemonCommand(args, {'targets': {}}, create_logger())
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'No targets available\n'
            'Use "add" command first\n'
        )
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_wrong_target(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=['foo'], conflicts='ignore', read_delay=0)

        command = DaemonCommand(
            args,
            {'targets': {'bar': {}}},
            create_logger(),
        )
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'Unknown target: foo\n'
        )
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_specific_target(self, INotifyRecursive, SyncWorker):
        INotifyRecursive.return_value = FakeINotify(
            events={
                Event(wd=1, mask=flags.CREATE, cookie=None, name="hello.txt"),
                Event(wd=2, mask=flags.CREATE, cookie=None, name="bar.txt"),
            },
            wd_map={
                1: '/home/jon/code/',
                2: '/home/jon/code/hoot',
            }
        )

        args = argparse.Namespace(targets=['foo'], conflicts='ignore', read_delay=0)
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/jon/code',
                    's3_uri': 's3://bucket/code',
                    'aws_secret_access_key': '23232323',
                    'aws_access_key_id': '########',
                    'region_name': 'eu-west-2',
                },
                'bar': {},
            }
        }
        command = DaemonCommand(args, config, create_logger())
        command.run(terminator=self.single_term)

        assert SyncWorker.call_count == 2
        assert INotifyRecursive.call_count == 1



# -*- coding: utf-8 -*-

import json

import mock

from s4 import utils


@mock.patch('getpass.getpass')
@mock.patch('builtins.input')
def test_get_input(getpass, input_fn):
    utils.get_input("give me some info", secret=False)

    assert getpass.call_count == 1
    assert input_fn.call_count == 0


@mock.patch('getpass.getpass')
@mock.patch('builtins.input')
def test_get_input_secret(getpass, input_fn):
    utils.get_input("give me some secret info", secret=True)

    assert getpass.call_count == 0
    assert input_fn.call_count == 1


class TestGetConfigFile(object):
    @mock.patch('s4.utils.CONFIG_FILE_PATH', '/i/dont/exist')
    def test_no_file(self):
        assert utils.get_config() == {'targets': {}}

    def test_correct_output(self, config_file):
        with open(config_file, 'w') as fp:
            json.dump({'local_folder': '/home/someone/something'}, fp)

        assert utils.get_config() == {'local_folder': '/home/someone/something'}



# -*- encoding: utf-8 -*-

import argparse

from s4.commands.ls_command import LsCommand
from tests.utils import (
    create_logger,
    get_timestamp,
    set_local_index,
    set_s3_index,
)


class TestLsCommand(object):
    def test_empty_config(self, capsys):
        args = argparse.Namespace(target='idontexist')
        command = LsCommand(args, {}, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'You have not added any targets yet\n'
            'Use the "add" command to do this\n'
        )

    def test_missing_target(self, capsys):
        args = argparse.Namespace(target='idontexist')
        config = {
            'targets': {
                'foo': {},
                'bar': {},
            }
        }
        command = LsCommand(args, config, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            '"idontexist" is an unknown target\n'
            'Choices are: [\'bar\', \'foo\']\n'
        )

    def test_correct_output_empty(self, s3_client, local_client, capsys):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }
        args = argparse.Namespace(target='foo', sort_by='key', descending=False)
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            'key    local    s3\n'
            '-----  -------  ----\n'
        )

    def test_correct_output_nonempty(self, s3_client, local_client, capsys):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }
        set_s3_index(s3_client, {
            'milk': {
                'local_timestamp': get_timestamp(1989, 10, 23, 11, 30)
            },
            'honey': {
                'local_timestamp': get_timestamp(2016, 12, 12, 8, 30)
            },
            'ginger': {
                'local_timestamp': None,
            }
        })
        set_local_index(local_client, {
            'milk': {
                'local_timestamp': get_timestamp(2016, 12, 12, 8, 30)
            },
            'honey': {
                'local_timestamp': get_timestamp(2016, 11, 10, 18, 40)
            },
            'lemon': {
                'local_timestamp': get_timestamp(2017, 2, 2, 8, 30)
            },
            'ginger': {
                'local_timestamp': None,
            }
        })

        args = argparse.Namespace(
            target='foo',
            sort_by='key',
            show_all=False,
            descending=False,
        )
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            'key    local                s3\n'
            '-----  -------------------  -------------------\n'
            'honey  2016-11-10 18:40:00  2016-12-12 08:30:00\n'
            'lemon  2017-02-02 08:30:00\n'
            'milk   2016-12-12 08:30:00  1989-10-23 11:30:00\n'

        )

    def test_show_all(self, s3_client, local_client, capsys):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }
        set_s3_index(s3_client, {
            'cheese': {
                'local_timestamp': get_timestamp(2017, 12, 12, 8, 30)
            },
            'crackers': {
                'local_timestamp': None,
            }
        })
        set_local_index(local_client, {
            'cheese': {
                'local_timestamp': get_timestamp(2017, 2, 2, 8, 30)
            },
            'crackers': {
                'local_timestamp': None,
            }
        })

        args = argparse.Namespace(
            target='foo',
            sort_by='key',
            show_all=True,
            descending=False,
        )
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            'key       local                s3\n'
            '--------  -------------------  -------------------\n'
            'cheese    2017-02-02 08:30:00  2017-12-12 08:30:00\n'
            'crackers  <deleted>\n'

        )



# -*- coding: utf-8 -*-

import logging
import traceback

from s4.clients import SyncState
from s4.resolution import Resolution


class SyncWorker(object):
    def __init__(
            self,
            client_1,
            client_2,
            start_callback=None,
            update_callback=None,
            complete_callback=None,
            action_callback=None,
            conflict_handler=None,
    ):
        self.client_1 = client_1
        self.client_2 = client_2
        self.logger = logging.getLogger(str(self))
        self.start_callback = start_callback
        self.update_callback = update_callback
        self.complete_callback = complete_callback
        self.action_callback = action_callback
        self.conflict_handler = conflict_handler

    def __repr__(self):
        return 'SyncWorker<{}, {}>'.format(self.client_1.get_uri(), self.client_2.get_uri())

    def sync(self, conflict_choice=None, keys=None, dry_run=False):
        self.client_1.lock()
        self.client_2.lock()
        try:
            resolutions, unhandled_events = self.get_sync_states(keys)

            self.logger.debug(
                'There are %s unhandled events for the user to solve', len(unhandled_events)
            )
            self.logger.debug(
                'There are %s automatically resolved calls', len(resolutions)
            )
            for key in sorted(unhandled_events.keys()):
                action_1, action_2 = unhandled_events[key]
                if conflict_choice == '1':
                    resolutions[key] = Resolution.get_resolution(
                        key, action_1, self.client_2, self.client_1
                    )
                elif conflict_choice == '2':
                    resolutions[key] = Resolution.get_resolution(
                        key, action_2, self.client_1, self.client_2
                    )
                if self.conflict_handler is not None:
                    resolution = self.conflict_handler(
                        key, action_1, self.client_1, action_2, self.client_2
                    )
                    if resolution is not None:
                        resolutions[key] = resolution
                    else:
                        self.logger.info('Ignoring sync conflict for %s', key)
                else:
                    self.logger.info('Unable to resolve conflict for %s', key)

            self.run_resolutions(resolutions, dry_run)

        finally:
            self.client_1.unlock()
            self.client_2.unlock()

    def get_sync_states(self, keys=None):
        # we store a list of resolutions to make sure we can handle everything before
        # running any updates on the file system and indexes
        resolutions = {}

        # list of unhandled events which cannot be solved automatically (or alternatively the
        # the automated solution has not yet been implemented)
        unhandled_events = {}

        self.logger.debug('Generating deferred calls based on client states')
        for key, state_1, state_2 in self.get_states(keys):
            self.logger.debug('%s: %s %s', key, state_1, state_2)
            if state_1.state == SyncState.NOCHANGES and state_2.state == SyncState.NOCHANGES:
                if state_1.remote_timestamp == state_2.remote_timestamp:
                    continue
                elif state_1.remote_timestamp > state_2.remote_timestamp:
                    resolutions[key] = Resolution(
                        Resolution.UPDATE, self.client_2, self.client_1,
                        key, state_1.remote_timestamp
                    )
                elif state_2.remote_timestamp > state_1.remote_timestamp:
                    resolutions[key] = Resolution(
                        Resolution.UPDATE, self.client_1, self.client_2,
                        key, state_2.remote_timestamp
                    )

            elif state_1.state == SyncState.CREATED and state_2.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_2, self.client_1, key, state_1.local_timestamp
                )

            elif state_2.state == SyncState.CREATED and state_1.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_1, self.client_2, key, state_2.local_timestamp
                )

            elif state_1.state == SyncState.NOCHANGES and state_2.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_2, self.client_1, key, state_1.remote_timestamp
                )

            elif state_2.state == SyncState.NOCHANGES and state_1.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_1, self.client_2, key, state_2.remote_timestamp
                )
            elif state_1.state == SyncState.UPDATED and state_2.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_2, self.client_1, key, state_1.local_timestamp
                )

            elif state_2.state == SyncState.UPDATED and state_1.state == SyncState.DOESNOTEXIST:
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_2, self.client_1, key, state_1.local_timestamp
                )

            elif (
                state_1.state in (SyncState.DELETED, SyncState.DOESNOTEXIST) and
                state_2.state in (SyncState.DELETED, SyncState.DOESNOTEXIST)
            ):
                # nothing to do, they have already both been deleted/do not exist
                continue

            elif (
                state_1.state == SyncState.UPDATED and
                state_2.state == SyncState.NOCHANGES and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.UPDATE, self.client_2, self.client_1, key, state_1.local_timestamp
                )

            elif (
                state_2.state == SyncState.UPDATED and
                state_1.state == SyncState.NOCHANGES and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.UPDATE, self.client_1, self.client_2, key, state_2.local_timestamp
                )

            elif (
                state_1.state == SyncState.DELETED and
                state_2.state == SyncState.NOCHANGES and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.DELETE, self.client_2, None, key, state_1.remote_timestamp
                )

            elif (
                state_2.state == SyncState.DELETED and
                state_1.state == SyncState.NOCHANGES and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.DELETE, self.client_1, None, key, state_2.remote_timestamp
                )

            elif (
                state_1.state == SyncState.DELETED and
                state_2.state == SyncState.CREATED and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_1, self.client_2, key, state_2.local_timestamp
                )

            elif (
                state_2.state == SyncState.DELETED and
                state_1.state == SyncState.CREATED and
                state_1.remote_timestamp == state_2.remote_timestamp
            ):
                resolutions[key] = Resolution(
                    Resolution.CREATE, self.client_2, self.client_1, key, state_1.local_timestamp
                )

            else:
                unhandled_events[key] = (state_1, state_2)

            self.logger.debug('Action=%s', resolutions.get(key))

        return resolutions, unhandled_events

    def run_resolutions(self, resolutions, dry_run=False):
        # call everything once we know we can handle all of it
        self.logger.debug('There are %s total deferred calls', len(resolutions))
        success = []
        try:
            for key in sorted(resolutions.keys()):
                resolution = resolutions[key]
                if self.action_callback is not None:
                    self.action_callback(resolution)

                if resolution.action == Resolution.UPDATE:
                    deferred_function = self.move_client
                elif resolution.action == Resolution.CREATE:
                    deferred_function = self.move_client
                elif resolution.action == Resolution.DELETE:
                    deferred_function = self.delete_client
                else:
                    raise ValueError('Unknown resolution', resolution)

                if dry_run:
                    continue

                try:
                    deferred_function(resolution)
                    self.client_1.update_index_entry(key)
                    self.client_2.update_index_entry(key)
                    success.append(key)
                except Exception as e:
                    self.logger.error('An error occurred while trying to update %s: %s', key, e)
                    self.logger.debug(traceback.format_exc())
        except KeyboardInterrupt:
            self.logger.warning('Session interrupted by Keyboard Interrupt. Cleaning up....')

        if len(success) > 0:
            self.logger.info('Flushing Index to Storage')
            self.client_1.flush_index()
            self.client_2.flush_index()
        else:
            self.logger.info('Nothing to update')

        return success

    def get_states(self, keys=None):
        client_1_actions = self.client_1.get_all_actions()
        client_2_actions = self.client_2.get_all_actions()

        all_keys = set(client_1_actions) | set(client_2_actions)
        self.logger.debug(
            '%s keys in total (%s for %s and %s for %s)',
            len(all_keys),
            len(client_1_actions), self.client_1.get_uri(),
            len(client_2_actions), self.client_2.get_uri()
        )
        if keys is None:
            target_keys = sorted(all_keys)
        else:
            target_keys = keys

        DOES_NOT_EXIST = SyncState(SyncState.DOESNOTEXIST, None, None)
        for key in target_keys:
            action_1 = client_1_actions.get(key, DOES_NOT_EXIST)
            action_2 = client_2_actions.get(key, DOES_NOT_EXIST)
            yield key, action_1, action_2

    def move_client(self, resolution):
        sync_object = resolution.from_client.get(resolution.key)

        if self.start_callback is not None:
            self.start_callback(sync_object)

        resolution.to_client.put(
            resolution.key,
            sync_object,
            callback=self.update_callback,
        )

        if self.complete_callback is not None:
            self.complete_callback(sync_object)

        resolution.to_client.set_remote_timestamp(resolution.key, resolution.timestamp)
        resolution.from_client.set_remote_timestamp(resolution.key, resolution.timestamp)

    def delete_client(self, resolution):
        resolution.to_client.delete(resolution.key)
        resolution.to_client.set_remote_timestamp(resolution.key, resolution.timestamp)



#! -*- encoding: utf-8 -*-

from datetime import datetime

from tabulate import tabulate

from s4.commands import Command


class LsCommand(Command):
    def run(self):
        if 'targets' not in self.config:
            self.logger.info('You have not added any targets yet')
            self.logger.info('Use the "add" command to do this')
            return
        if self.args.target not in self.config['targets']:
            all_targets = sorted(list(self.config['targets'].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info('Choices are: %s', all_targets)
            return

        target = self.config['targets'][self.args.target]
        client_1, client_2 = self.get_clients(target)

        sort_by = self.args.sort_by.lower()
        descending = self.args.descending

        keys = set(client_1.index) | set(client_2.index)

        data = []
        for key in sorted(keys):
            entry_1 = client_1.index.get(key, {})
            entry_2 = client_2.index.get(key, {})

            ts_1 = entry_1.get('local_timestamp')
            ts_2 = entry_2.get('local_timestamp')

            if self.args.show_all or ts_1 is not None:
                data.append((
                    key,
                    datetime.utcfromtimestamp(int(ts_1)) if ts_1 is not None else '<deleted>',
                    datetime.utcfromtimestamp(int(ts_2)) if ts_2 is not None else None,
                ))

        headers = ['key', 'local', 's3']
        data = sorted(data, reverse=descending, key=lambda x: x[headers.index(sort_by)])

        print(tabulate(data, headers=headers))



#! -*- encoding: utf8 -*
import os

from s4 import utils
from s4.commands import Command


class AddCommand(Command):
    def run(self):
        target = self.args.copy_target_credentials
        all_targets = list(self.config['targets'].keys())
        if target is not None and target not in all_targets:
            self.logger.info('"%s" is an unknown target', target)
            self.logger.info('Choices are: %s', all_targets)
            return

        entry = {}
        entry['local_folder'] = os.path.expanduser(utils.get_input('local folder: '))
        entry['s3_uri'] = utils.get_input('s3 uri: ')
        entry['region_name'] = utils.get_input('region name: ')

        if target is not None:
            entry['aws_access_key_id'] = self.config['targets'][target]['aws_access_key_id']
            entry['aws_secret_access_key'] = self.config['targets'][target]['aws_secret_access_key']
        else:
            entry['aws_access_key_id'] = utils.get_input('AWS Access Key ID: ')
            entry['aws_secret_access_key'] = utils.get_input('AWS Secret Access Key: ', secret=True)

        default_name = os.path.basename(entry['s3_uri'])
        name = utils.get_input('Provide a name for this entry [{}]: '.format(default_name))

        if not name:
            name = default_name

        if 'targets' not in self.config:
            self.config['targets'] = {}

        self.config['targets'][name] = entry

        utils.set_config(self.config)



#! -*- encoding: utf-8 -*-
from collections import defaultdict

# Dont crash on import if the underlying operating system does not support INotify
try:
    from inotify_simple import flags
    from s4.inotify_recursive import INotifyRecursive
    supported = True
except OSError:
    supported = False

from s4.commands import Command


class DaemonCommand(Command):
    def run(self, terminator=lambda x: False):
        if not supported:
            self.logger.info('Cannot run INotify on your operating system')
            self.logger.info('Only Linux machines are officially supported for this command')
            return

        all_targets = list(self.config['targets'].keys())
        if not self.args.targets:
            targets = all_targets
        else:
            targets = self.args.targets

        if not targets:
            self.logger.info('No targets available')
            self.logger.info('Use "add" command first')
            return

        for target in targets:
            if target not in self.config['targets']:
                self.logger.info("Unknown target: %s", target)
                return

        notifier = INotifyRecursive()
        watch_flags = flags.CREATE | flags.DELETE | flags.MODIFY

        watch_map = {}

        for target in targets:
            entry = self.config['targets'][target]
            path = entry['local_folder']
            self.logger.info("Watching %s", path)
            for wd in notifier.add_watches(path.encode('utf8'), watch_flags):
                watch_map[wd] = target

            # Check for any pending changes
            worker = self.get_sync_worker(target)
            worker.sync(conflict_choice=self.args.conflicts)

        index = 0
        while not terminator(index):
            index += 1

            to_run = defaultdict(set)
            for event in notifier.read(read_delay=self.args.read_delay):
                target = watch_map[event.wd]

                # Dont bother running for .index
                if event.name not in ('.index', '.s4lock'):
                    to_run[target].add(event.name)

            for target, keys in to_run.items():
                worker = self.get_sync_worker(target)

                # Should ideally be setting keys to sync
                self.logger.info('Syncing {}'.format(worker))
                worker.sync(conflict_choice=self.args.conflicts)



#! -*- encoding: utf8 -*-
from inotify_simple import flags

import pytest

from s4.inotify_recursive import INotifyRecursive


class TestINotifyRecursive(object):
    @pytest.mark.timeout(5)
    def test_add_watches(self, tmpdir):
        foo = tmpdir.mkdir("foo")
        bar = tmpdir.mkdir("bar")
        baz = bar.mkdir("baz")

        notifier = INotifyRecursive()
        result_1 = notifier.add_watches(str(foo), flags.CREATE)
        result_2 = notifier.add_watches(str(bar), flags.CREATE)

        assert sorted(result_1.values()) == sorted([str(foo)])
        assert sorted(result_2.values()) == sorted([str(bar), str(baz)])

        bar.join("hello.txt").write("hello")
        foo.join("fennek.md").write("*jumps*")
        baz.mkdir("bong")

        events = notifier.read()
        assert len(events) == 3

        assert events[0].name == 'hello.txt'
        assert result_2[events[0].wd] == str(bar)

        assert events[1].name == 'fennek.md'
        assert result_1[events[1].wd] == str(foo)

        assert events[2].name == 'bong'
        assert result_2[events[2].wd] == str(baz)



# -*- encoding: utf-8 -*-

import argparse
import json

from s4.commands.rm_command import RmCommand
from tests.utils import create_logger


class TestRemoveCommand(object):
    def test_empty(self, capsys, config_file):
        args = argparse.Namespace(target='foo')
        command = RmCommand(args, {}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'You have not added any targets yet\n'
        )

    def test_missing(self, capsys, config_file):
        args = argparse.Namespace(target='foo')
        command = RmCommand(args, {
            'targets': {
                'bar': {}
            }
        }, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            '"foo" is an unknown target\n'
            'Choices are: [\'bar\']\n'
        )

    def test_remove_target(self, capsys, config_file):
        args = argparse.Namespace(target='foo')
        command = RmCommand(args, {
            'targets': {
                'bar': {},
                'foo': {},
            }
        }, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == err == ''

        with open(config_file, 'rt') as fp:
            new_config = json.load(fp)

        expected_config = {'targets': {'bar': {}}}
        assert new_config == expected_config



# -*- encoding: utf-8 -*-

import pytest

from s4.clients import SyncState, local, s3
from s4.resolution import Resolution


class TestResolution(object):
    def test_equal_wrong_instance(self):
        resolution = Resolution(Resolution.CREATE, None, None, 'bar', 23232)
        assert resolution != "Not a Resolution"

    def test_equal_to_self(self):
        resolution = Resolution(Resolution.UPDATE, None, None, 'fffff', 232323)
        assert resolution == resolution

    def test_equal(self):
        resolution_1 = Resolution(Resolution.UPDATE, None, None, 'fffff', 232323)
        resolution_2 = Resolution(Resolution.UPDATE, None, None, 'fffff', 232323)
        assert resolution_1 == resolution_2

    def test_not_equal(self):
        resolution_1 = Resolution(Resolution.UPDATE, None, None, 'fffff', 232323)
        resolution_2 = Resolution(Resolution.DELETE, None, None, 'wew', 3823)
        assert resolution_1 != resolution_2

    def test_repr(self):
        s3_client = s3.S3SyncClient(None, 'mortybucket', 'dimensional/portals')
        local_client = local.LocalSyncClient('/home/picklerick')
        resolution = Resolution(Resolution.CREATE, s3_client, local_client, 'foo', 20023)
        expected_repr = (
            "Resolution<action=CREATE, "
            "to=s3://mortybucket/dimensional/portals/, "
            "from=/home/picklerick/, "
            "key=foo, timestamp=20023>"
        )
        assert repr(resolution) == expected_repr

    @pytest.mark.parametrize(['state', 'action'], [
        (SyncState.UPDATED, Resolution.UPDATE),
        (SyncState.UPDATED, Resolution.UPDATE),
        (SyncState.CREATED, Resolution.CREATE),
        (SyncState.DELETED, Resolution.DELETE),
    ])
    def test_get_resolution_updated(self, state, action, s3_client, local_client):
        sync_state = SyncState(state, 1234, 4567)
        resolution = Resolution.get_resolution(
            'foo/bar', sync_state, s3_client, local_client
        )
        assert resolution.action == action
        assert resolution.to_client == s3_client
        assert resolution.key == 'foo/bar'
        if state != SyncState.DELETED:
            assert resolution.timestamp == 1234
            assert resolution.from_client == local_client
        else:
            assert resolution.timestamp == 4567
            assert resolution.from_client is None

    def test_invalid_sync_state(self):
        with pytest.raises(ValueError) as exc:
            Resolution.get_resolution('', SyncState('Invalid', None, None), None, None)
        assert exc.value



# -*- coding: utf-8 -*-

import mock
import pytest

from s4 import cli


# TODO: Should catch KeyboardExceptions and raise them again
class TestMain(object):

    @mock.patch('argparse.ArgumentParser.print_help')
    def test_no_arguments_prints_help(self, print_help):
        cli.main([])
        assert print_help.call_count == 1

    @pytest.mark.parametrize(['loglevel'], [('INFO', ), ('DEBUG', )])
    @mock.patch('logging.basicConfig')
    def test_timestamps(self, basicConfig, loglevel):
        cli.main(['--timestamps', '--log-level', loglevel, 'version'])
        assert basicConfig.call_args[1]['format'].startswith('%(asctime)s: ')

    @mock.patch('logging.basicConfig')
    def test_debug_loglevel(self, basicConfig):
        cli.main(['--log-level=DEBUG', 'version'])
        assert basicConfig.call_args[1]['format'].startswith('%(levelname)s:%(module)s')
        assert basicConfig.call_args[1]['level'] == 'DEBUG'

    def test_version_command(self, capsys):
        cli.main(['version'])
        out, err = capsys.readouterr()
        assert out == '{}\n'.format(cli.VERSION)

    @mock.patch('s4.cli.LsCommand')
    def test_ls_command(self, LsCommand):
        cli.main(['ls', 'foo'])
        assert LsCommand.call_count == 1

    @mock.patch('s4.cli.DaemonCommand')
    def test_daemon_command(self, DaemonCommand):
        cli.main(['daemon'])
        assert DaemonCommand.call_count == 1

    @mock.patch('s4.cli.SyncCommand')
    def test_sync_command(self, SyncCommand):
        cli.main(['sync', 'foo'])
        assert SyncCommand.call_count == 1

    @mock.patch('s4.cli.EditCommand')
    def test_edit_command(self, EditCommand):
        cli.main(['edit', 'foo'])
        assert EditCommand.call_count == 1

    @mock.patch('s4.cli.TargetsCommand')
    def test_targets_command(self, TargetsCommand):
        cli.main(['targets'])
        assert TargetsCommand.call_count == 1

    @mock.patch('s4.cli.RmCommand')
    def test_rm_command(self, RmCommand):
        cli.main(['rm', 'foo'])
        assert RmCommand.call_count == 1

    @mock.patch('s4.cli.AddCommand')
    def test_add_command(self, AddCommand):
        cli.main(['add'])
        assert AddCommand.call_count == 1



#! -*- encoding: utf8 -*-

from s4 import sync
from s4.clients.local import get_local_client
from s4.clients.s3 import get_s3_client


class Command(object):
    def __init__(self, args, config, logger):
        self.args = args
        self.config = config
        self.logger = logger

    def get_sync_worker(self, target):
        entry = self.config['targets'][target]
        client_1, client_2 = self.get_clients(entry)
        return sync.SyncWorker(client_1, client_2)

    def get_clients(self, entry):
        target_1 = entry['local_folder']
        target_2 = entry['s3_uri']
        aws_access_key_id = entry['aws_access_key_id']
        aws_secret_access_key = entry['aws_secret_access_key']
        region_name = entry['region_name']

        # append trailing slashes to prevent incorrect prefix matching on s3
        if not target_1.endswith('/'):
            target_1 += '/'
        if not target_2.endswith('/'):
            target_2 += '/'

        client_1 = get_local_client(target_1)
        client_2 = get_s3_client(target_2, aws_access_key_id, aws_secret_access_key, region_name)
        return client_1, client_2



# -*- coding: utf-8 -*-

import gzip
import io
import json
import os
import shutil
import tempfile

import filelock
import mock
import pytest

from s4.clients import SyncObject, local
from tests import utils


class TestTraverse(object):
    def setup_method(self):
        self.target_folder = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.target_folder)

    def test_non_existent_folder(self):
        assert list(local.traverse('/i/definetely/do/not/exist')) == []

    def test_empty_folder(self):
        assert list(local.traverse(self.target_folder)) == []

    def test_correct_output(self):
        items = [
            'baz/zoo',
            'foo',
            'bar.md',
            'baz/bar',
            '.index',
            'garbage~',
            'saw/.index',
            'SomeProject/.git/2aa58a13dcbca4b13a244dadf5536865ead5e1',
            'SomeProject/hello.py',
            'SomeProject/hello.pyo',
            'SomeProject/hello.pyc',
        ]
        for item in items:
            utils.write_local(os.path.join(self.target_folder, item))

        actual_output = list(local.traverse(
            self.target_folder,
            ignore_files={'.index', '.idontexist', '*~', '.git', '*.py[co]'}
        ))

        expected_output = ['bar.md', 'baz/bar', 'baz/zoo', 'foo', 'SomeProject/hello.py']
        assert sorted(actual_output) == sorted(expected_output)


class TestLocalSyncClient(object):
    def test_get_client_name(self, local_client):
        assert local_client.get_client_name() == 'local'

    def test_repr(self):
        client = local.LocalSyncClient('/my/test/path')
        assert repr(client) == 'LocalSyncClient</my/test/path>'

    def test_lock(self, local_client):
        local_client2 = local.LocalSyncClient(local_client.path)
        local_client.lock(timeout=0.01)
        with pytest.raises(filelock.Timeout):
            local_client2.lock(timeout=0.01)
        local_client.unlock()

        local_client.lock(timeout=0.01)
        local_client2.unlock()

    def test_put_callback(self, local_client):
        mock_callback = mock.MagicMock()

        data = b'hello'
        local_client.put(
            key='solair/astora',
            sync_object=SyncObject(io.BytesIO(data), len(data), 7194),
            callback=mock_callback,
        )

        assert mock_callback.call_count == 1
        mock_callback.assert_called_with(5)

    def test_put_new(self, local_client):
        data = b'hi'
        local_client.put(
            key='foo/hello_world.txt',
            sync_object=SyncObject(io.BytesIO(data), len(data), 20000)
        )

        assert local_client.index['foo/hello_world.txt']['remote_timestamp'] == 20000
        assert utils.get_local_contents(local_client, 'foo/hello_world.txt') == data

    def test_get_uri(self):
        client = local.LocalSyncClient('/home/michael')
        assert client.get_uri() == '/home/michael/'
        assert client.get_uri('banana.txt') == '/home/michael/banana.txt'

    def test_put_existing(self, local_client):
        utils.set_local_index(local_client, {
            'doge.txt': {'local_timestamp': 1111111}
        })

        data = b'canis lupus familiaris'
        local_client.put(
            key='doge.txt',
            sync_object=SyncObject(io.BytesIO(data), len(data), 20000)
        )

        assert local_client.index['doge.txt']['remote_timestamp'] == 20000
        assert local_client.index['doge.txt']['local_timestamp'] == 1111111

        assert utils.get_local_contents(local_client, 'doge.txt') == data

    def test_get_existing(self, local_client):
        data = b'blue green yellow'
        utils.set_local_contents(local_client, 'whatup.md', data=data.decode('utf8'))
        sync_object = local_client.get('whatup.md')
        assert sync_object.fp.read() == data
        assert sync_object.total_size == len(data)

    def test_get_non_existant(self, local_client):
        assert local_client.get('idontexist.md') is None

    def test_delete_existing(self, local_client):
        target_file = os.path.join(local_client.path, 'foo')
        utils.set_local_contents(local_client, 'foo', 222222)

        assert os.path.exists(target_file) is True
        assert local_client.delete('foo') is True
        assert os.path.exists(target_file) is False

    def test_delete_non_existant(self, local_client):
        assert local_client.delete('idontexist.txt') is False

    def test_index_path(self):
        client = local.LocalSyncClient('/hello/from/the/magic/tavern')
        assert client.index_path() == '/hello/from/the/magic/tavern/.index'

    def test_load_index(self, local_client):
        data = {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            },
            'bar/baz.txt': {
                'local_timestamp': 5000,
                'remote_timestamp': 5000,
            },
        }
        utils.set_local_index(local_client, data)

        assert local_client.index == data

    def test_get_local_keys(self, local_client):
        utils.set_local_index(local_client, {})  # .index file should not come up in results
        utils.set_local_contents(local_client, '.bashrc')
        utils.set_local_contents(local_client, 'foo')
        utils.set_local_contents(local_client, 'bar')

        actual_output = local_client.get_local_keys()
        expected_output = ['foo', 'bar', '.bashrc']
        assert sorted(expected_output) == sorted(actual_output)

    def test_get_index_keys(self, local_client):
        data = {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            },
            'bar/baz.txt': {
                'local_timestamp': 5000,
                'remote_timestamp': 5000,
            },
        }
        utils.set_local_index(local_client, data)

        actual_output = local_client.get_index_keys()
        expected_output = ['foo', 'bar/baz.txt']
        assert sorted(list(actual_output)) == sorted(expected_output)

    def test_all_keys(self, local_client):
        utils.set_local_contents(local_client, '.index')
        utils.set_local_contents(local_client, 'foo')
        utils.set_local_contents(local_client, 'bar/boo.md')
        data = {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            },
            'bar/baz.txt': {
                'local_timestamp': 5000,
                'remote_timestamp': 5000,
            },
        }
        utils.set_local_index(local_client, data)

        actual_output = local_client.get_all_keys()
        expected_output = ['foo', 'bar/boo.md', 'bar/baz.txt']
        assert sorted(actual_output) == sorted(expected_output)

    def test_update_index_empty(self, local_client):
        utils.set_local_contents(local_client, 'foo', 13371337)
        utils.set_local_contents(local_client, 'bar', 50032003)

        local_client.update_index()
        # remote timestamp should not be included since it does not exist
        expected_output = {
            'foo': {
                'local_timestamp': 13371337,
                'remote_timestamp': None,
            },
            'bar': {
                'local_timestamp': 50032003,
                'remote_timestamp': None,
            },
        }
        assert local_client.index == expected_output

    def test_update_index_non_empty(self, local_client):
        utils.set_local_contents(local_client, 'foo', 13371337)
        utils.set_local_contents(local_client, 'bar', 50032003)

        utils.set_local_index(local_client, {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            },
            'bar': {
                'local_timestamp': 5000,
                'remote_timestamp': 5000,
            },
            'baz': {
                'local_timestamp': 5000,
                'remote_timestamp': 5000,
            }
        })

        local_client.update_index()
        expected_output = {
            'foo': {
                'local_timestamp': 13371337,
                'remote_timestamp': 4000,
            },
            'bar': {
                'local_timestamp': 50032003,
                'remote_timestamp': 5000,
            },
            'baz': {
                'local_timestamp': None,
                'remote_timestamp': 5000,
            },
        }
        assert local_client.index == expected_output

    def test_get_real_local_timestamp(self, local_client):
        utils.set_local_contents(local_client, 'atcg', 2323230)

        assert local_client.get_real_local_timestamp('atcg') == 2323230
        assert local_client.get_real_local_timestamp('dontexist') is None

    def test_get_all_real_local_timestamps(self, local_client):
        utils.set_local_contents(local_client, 'red', 2323230)
        utils.set_local_contents(local_client, 'blue', 80808008)

        expected_output = {
            'red': 2323230,
            'blue': 80808008,
        }
        actual_output = local_client.get_all_real_local_timestamps()
        assert actual_output == expected_output

    def test_get_index_timestamps(self, local_client):
        utils.set_local_index(local_client, {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 32000,
            },
            'bar': {
                'local_timestamp': 9999999,
            }
        })
        assert local_client.get_index_local_timestamp('foo') == 4000
        assert local_client.get_index_local_timestamp('dontexist') is None
        assert local_client.get_index_local_timestamp('bar') == 9999999

        assert local_client.get_remote_timestamp('foo') == 32000
        assert local_client.get_remote_timestamp('dontexist') is None
        assert local_client.get_remote_timestamp('bar') is None

    def test_set_index_local_timestamp(self, local_client):
        utils.set_local_index(local_client, {
            'foo': {
                'local_timestamp': 100000,
                'remote_timestamp': 2222222222,
            }
        })
        local_client.set_index_local_timestamp('foo', 123456)
        assert local_client.get_index_local_timestamp('foo') == 123456

    def test_set_index_local_timestamp_non_existent(self, local_client):
        local_client.set_index_local_timestamp('foo', 222222)
        assert local_client.get_index_local_timestamp('foo') == 222222

    def test_get_all_index_local_timestamps(self, local_client):
        utils.set_local_index(local_client, {
            'frap': {
                'local_timestamp': 4000,
            },
            'brap': {
                'local_timestamp': 9999999,
            }
        })

        expected_output = {
            'frap': 4000,
            'brap': 9999999,
        }
        actual_output = local_client.get_all_index_local_timestamps()
        assert actual_output == expected_output

    def test_get_all_remote_timestamps(self, local_client):
        utils.set_local_index(local_client, {
            'frap': {
                'remote_timestamp': 4000,
            },
            'brap': {
                'remote_timestamp': 9999999,
            }
        })

        expected_output = {
            'frap': 4000,
            'brap': 9999999,
        }
        actual_output = local_client.get_all_remote_timestamps()
        assert actual_output == expected_output

    def test_interrupted_put(self, local_client):
        utils.set_local_contents(local_client, 'keychain', data='iamsomedatathatexists')

        sync_object = SyncObject(utils.InterruptedBytesIO(), 900000, 3000)

        with pytest.raises(ValueError):
            local_client.put('keychain', sync_object)

        result = local_client.get('keychain')
        assert result.fp.read() == b'iamsomedatathatexists'

    @pytest.mark.parametrize(['compressed', 'method'], [
        (True, gzip.open),
        (False, open),
    ])
    def test_flush_index(self, compressed, method, local_client):
        target_index = {
            'foo': {
                'local_timestamp': 4000,
                'remote_timestamp': 6000,
            }
        }

        local_client.index = target_index
        local_client.flush_index(compressed=compressed)

        with method(local_client.index_path(), 'rt') as fp:
            index = json.load(fp)

        assert index == target_index

    def test_interrupted_flush_index(self, local_client):
        target_index = {
            'red': {
                'local_timestamp': 4000,
                'remote_timestamp': 4000,
            }
        }

        utils.set_local_index(local_client, target_index)

        local_client.index = {'invalid_data': []}

        with mock.patch('json.dump') as json_dump:
            json_dump.side_effect = ValueError('something went wrong')
            with pytest.raises(ValueError):
                local_client.flush_index(compressed=False)

        with open(local_client.index_path(), 'rt') as fp:
            index = json.load(fp)

        assert index == target_index

    def test_ignore_files(self, local_client):
        utils.set_local_contents(local_client, '.syncignore', timestamp=3200, data=(
            '*.zip\n'
            'foo*\n'
        ))
        local_client.reload_ignore_files()

        utils.set_local_index(local_client, {
            'pony.tar': {
                'local_timestamp': 4000,
                'remote_timestamp': 3000,
            }
        })

        utils.set_local_contents(local_client, 'test.zip')
        utils.set_local_contents(local_client, 'foobar')
        utils.set_local_contents(local_client, 'foo')
        utils.set_local_contents(local_client, 'pony.tar', timestamp=8000)

        local_client.update_index()

        expected_index = {
            '.syncignore': {
                'local_timestamp': 3200,
                'remote_timestamp': None,
            },
            'pony.tar': {
                'local_timestamp': 8000,
                'remote_timestamp': 3000,
            }
        }
        assert local_client.index == expected_index



# -*- coding: utf-8 -*-

import fnmatch
import gzip
import json
import logging
import os
import shutil
import tempfile

# Use the built-in version of scandir/walk if possible, otherwise
# use the scandir module version
try:
    from os import scandir
except ImportError:
    from scandir import scandir

import filelock

import magic

from s4.clients import SyncClient, SyncObject

logger = logging.getLogger(__name__)


def get_local_client(target):
    return LocalSyncClient(target)


def traverse(path, ignore_files=None):
    if not os.path.exists(path):
        return

    if ignore_files is None:
        ignore_files = []

    for item in scandir(path):
        if any(fnmatch.fnmatch(item.name, pattern) for pattern in ignore_files):
            logger.debug('Ignoring %s', item)
            continue

        if item.is_dir():
            for result in traverse(item.path, ignore_files):
                yield os.path.join(item.name, result)
        else:
            yield item.name


class LocalSyncClient(SyncClient):
    DEFAULT_IGNORE_FILES = ['.index', '.s4lock']
    LOCK_FILE_NAME = '.s4lock'

    def __init__(self, path):
        self.path = path
        self.reload_index()
        self.reload_ignore_files()
        self._lock = filelock.FileLock(self.lock_file)

    @property
    def lock_file(self):
        return self.get_uri(self.LOCK_FILE_NAME)

    def ensure_path(self, path):
        parent = os.path.dirname(path)
        if not os.path.exists(parent):
            os.makedirs(parent)

    def lock(self, timeout=10):
        """
        Advisory lock.
        Use to ensure that only one LocalSyncClient is working on the Target at the same time.
        """
        logger.debug("Locking %s", self.lock_file)
        if not os.path.exists(self.lock_file):
            self.ensure_path(self.lock_file)
            with open(self.lock_file, 'w'):
                os.utime(self.lock_file)
        self._lock.acquire(timeout=timeout)

    def unlock(self):
        """
        Unlock the active advisory lock.
        """
        logger.debug("Releasing lock %s", self.lock_file)
        self._lock.release()
        try:
            os.unlink(self.lock_file)
        except FileNotFoundError:
            pass

    def get_client_name(self):
        return 'local'

    def __repr__(self):
        return 'LocalSyncClient<{}>'.format(self.path)

    def get_uri(self, key=''):
        return os.path.join(self.path, key)

    def index_path(self):
        return os.path.join(self.path, '.index')

    def put(self, key, sync_object, callback=None):
        path = os.path.join(self.path, key)
        self.ensure_path(path)

        BUFFER_SIZE = 4096
        fd, temp_path = tempfile.mkstemp()

        try:
            with open(temp_path, 'wb') as fp1:
                while True:
                    data = sync_object.fp.read(BUFFER_SIZE)
                    fp1.write(data)
                    if callback is not None:
                        callback(len(data))
                    if len(data) < BUFFER_SIZE:
                        break
            shutil.move(temp_path, path)
        except Exception:
            os.remove(temp_path)
            raise
        finally:
            os.close(fd)

        self.set_remote_timestamp(key, sync_object.timestamp)

    def get(self, key):
        path = os.path.join(self.path, key)
        if os.path.exists(path):
            fp = open(path, 'rb')
            stat = os.stat(path)
            return SyncObject(fp, stat.st_size, stat.st_mtime)
        else:
            return None

    def delete(self, key):
        path = os.path.join(self.path, key)
        if os.path.exists(path):
            os.remove(path)
            return True
        else:
            return False

    def reload_index(self):
        self.index = self._load_index()

    def _load_index(self):
        index_path = self.index_path()
        if not os.path.exists(index_path):
            return {}

        content_type = magic.from_file(index_path, mime=True)
        if content_type == 'text/plain':
            logger.debug('Detected plaintext encoding for reading index')
            method = open
        elif content_type in ('application/gzip', 'application/x-gzip'):
            logger.debug('Detected gzip encoding for reading index')
            method = gzip.open
        else:
            raise ValueError('Index is of unknown type', content_type)

        with method(index_path, 'rt') as fp:
            data = json.load(fp)
        return data

    def flush_index(self, compressed=True):
        if compressed:
            logger.debug('Using gzip encoding for writing index')
            method = gzip.open
        else:
            logger.debug('Using plaintext encoding for writing index')
            method = open

        fd, temp_path = tempfile.mkstemp()
        with method(temp_path, 'wt') as fp:
            json.dump(self.index, fp)

        os.close(fd)

        shutil.move(temp_path, self.index_path())

    def get_local_keys(self):
        return list(traverse(self.path, ignore_files=self.ignore_files))

    def get_real_local_timestamp(self, key):
        full_path = os.path.join(self.path, key)
        if os.path.exists(full_path):
            return os.path.getmtime(full_path)
        else:
            return None

    def get_index_keys(self):
        return self.index.keys()

    def get_index_local_timestamp(self, key):
        return self.index.get(key, {}).get('local_timestamp')

    def get_all_real_local_timestamps(self):
        result = {}
        for key in self.get_local_keys():
            result[key] = self.get_real_local_timestamp(key)
        return result

    def get_all_remote_timestamps(self):
        return {key: value.get('remote_timestamp') for key, value in self.index.items()}

    def get_all_index_local_timestamps(self):
        return {key: value.get('local_timestamp') for key, value in self.index.items()}

    def set_index_local_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]['local_timestamp'] = timestamp

    def get_remote_timestamp(self, key):
        return self.index.get(key, {}).get('remote_timestamp')

    def set_remote_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]['remote_timestamp'] = timestamp

    def reload_ignore_files(self):
        ignore_path = os.path.join(self.path, '.syncignore')

        if os.path.exists(ignore_path):
            with open(ignore_path, 'r') as fp:
                ignore_list = fp.read().split('\n')
        else:
            ignore_list = []

        self.ignore_files = self.DEFAULT_IGNORE_FILES + ignore_list



#! -*- encoding: utf-8 -*-
import argparse

import mock

from s4.clients import SyncState
from s4.commands.sync_command import (
    SyncCommand,
    display_progress_bar,
    handle_conflict,
    hide_progress_bar,
    update_progress_bar,
)
from s4.resolution import Resolution
from s4.sync import SyncWorker

from tests.utils import FakeInputStream, create_logger, set_local_contents


@mock.patch('s4.utils.get_input')
class TestHandleConflict(object):
    def test_first_choice(self, get_input, s3_client, local_client):
        get_input.return_value = '1'

        action_1 = SyncState(
            SyncState.UPDATED,
            1111, 2222
        )
        action_2 = SyncState(
            SyncState.DELETED,
            3333, 4444
        )

        result = handle_conflict(
            'movie',
            action_1, s3_client,
            action_2, local_client,
        )
        assert result.action == Resolution.UPDATE

    def test_second_choice(self, get_input, s3_client, local_client):
        get_input.return_value = '2'

        action_1 = SyncState(
            SyncState.UPDATED,
            1111, 2222
        )
        action_2 = SyncState(
            SyncState.DELETED,
            3333, 4444
        )

        result = handle_conflict(
            'movie',
            action_1, s3_client,
            action_2, local_client,
        )
        assert result.action == Resolution.DELETE

    def test_skip(self, get_input, s3_client, local_client):
        get_input.return_value = 'X'

        action_1 = SyncState(
            SyncState.UPDATED,
            1111, 2222
        )
        action_2 = SyncState(
            SyncState.DELETED,
            3333, 4444
        )

        result = handle_conflict(
            'movie',
            action_1, s3_client,
            action_2, local_client,
        )
        assert result is None

    @mock.patch('s4.commands.sync_command.show_diff')
    def test_diff(self, show_diff, get_input, s3_client, local_client):
        get_input.side_effect = FakeInputStream([
            'd',
            'X',
        ])

        action_1 = SyncState(
            SyncState.UPDATED,
            1111, 2222
        )
        action_2 = SyncState(
            SyncState.DELETED,
            3333, 4444
        )

        result = handle_conflict(
            'movie',
            action_1, s3_client,
            action_2, local_client,
        )
        assert result is None
        assert show_diff.call_count == 1
        show_diff.assert_called_with(s3_client, local_client, 'movie')


def test_progressbar_smoketest(s3_client, local_client):
    # Just test that nothing blows up
    set_local_contents(
        local_client, 'history.txt',
        data='a long long time ago',
        timestamp=5000,
    )

    worker = SyncWorker(
        s3_client,
        local_client,
        start_callback=display_progress_bar,
        update_callback=update_progress_bar,
        complete_callback=hide_progress_bar,
    )
    worker.sync()


@mock.patch('s4.sync.SyncWorker')
class TestSyncCommand(object):
    def test_no_targets(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        command = SyncCommand(args, {'targets': {}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == err == ''
        assert SyncWorker.call_count == 0

    def test_wrong_target(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=['foo', 'bar'], conflicts=None, dry_run=False)
        command = SyncCommand(args, {'targets': {'baz': {}}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            '"bar" is an unknown target. Choices are: [\'baz\']\n'
            '"foo" is an unknown target. Choices are: [\'baz\']\n'
        )
        assert SyncWorker.call_count == 0

    def test_sync_error(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False, log_level="INFO")
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/mike/docs',
                    's3_uri': 's3://foobar/docs',
                    'aws_access_key_id': '3223323',
                    'aws_secret_access_key': '23#@423#@',
                    'region_name': 'us-east-1',
                },
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }
        SyncWorker.side_effect = ValueError('something bad happened')

        command = SyncCommand(args,  config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 2
        assert out == ''
        assert err == (
            "There was an error syncing 'bar': something bad happened\n"
            "There was an error syncing 'foo': something bad happened\n"
        )

    def test_sync_error_debug(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False, log_level="DEBUG")
        config = {
            'targets': {
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }
        SyncWorker.side_effect = ValueError('something bad happened')

        command = SyncCommand(args,  config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 1
        assert out == ''
        assert err.split('\n')[:2] == [
            "something bad happened",
            "Traceback (most recent call last):",
        ]

    def test_keyboard_interrupt(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/mike/docs',
                    's3_uri': 's3://foobar/docs',
                    'aws_access_key_id': '3223323',
                    'aws_secret_access_key': '23#@423#@',
                    'region_name': 'us-east-1',
                },
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }
        SyncWorker.side_effect = KeyboardInterrupt

        command = SyncCommand(args,  config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 1
        assert out == ''
        assert err == (
            'Quitting due to Keyboard Interrupt...\n'
        )

    def test_all_targets(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/mike/docs',
                    's3_uri': 's3://foobar/docs',
                    'aws_access_key_id': '3223323',
                    'aws_secret_access_key': '23#@423#@',
                    'region_name': 'us-east-1',
                },
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }

        command = SyncCommand(args,  config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ''
        assert err == (
            'Syncing bar [/home/mike/barmil/ <=> s3://foobar/barrel/]\n'
            'Syncing foo [/home/mike/docs/ <=> s3://foobar/docs/]\n'
        )
        assert SyncWorker.call_count == 2



#! -*- encoding: utf8 -*-

from s4.commands import Command


class TargetsCommand(Command):
    def run(self):
        for name in sorted(self.config['targets']):
            entry = self.config['targets'][name]
            print('{}: [{} <=> {}]'.format(
                name, entry['local_folder'], entry['s3_uri']
            ))



#! -*- encoding: utf-8 -*-

from s4 import utils
from s4.commands import Command


class RmCommand(Command):
    def run(self):
        if 'targets' not in self.config:
            self.logger.info('You have not added any targets yet')
            return
        if self.args.target not in self.config['targets']:
            all_targets = sorted(list(self.config['targets'].keys()))
            self.logger.info('"%s" is an unknown target', self.args.target)
            self.logger.info('Choices are: %s', all_targets)
            return

        del self.config['targets'][self.args.target]
        utils.set_config(self.config)



#! -*- encoding: utf8 -*-

try:
    from os import scandir
except ImportError:
    from scandir import scandir

from inotify_simple import INotify


class INotifyRecursive(INotify):
    def add_watches(self, path, mask):
        results = {}
        results[self.add_watch(path, mask)] = path

        for item in scandir(path):
            if item.is_dir():
                results.update(self.add_watches(item.path, mask))

        return results
